import os
from datetime import datetime
import time
import json
import csv
import boto3

import dyndb

class DynamoImporter():
    def __init__(self, table_name, region):
        self.resource = boto3.resource('dynamodb', region_name=region)
        self.table_name = table_name
        self.region = region
        self.table = self.resource.Table(table_name)

        self.keys = []

        for item in self.table.attribute_definitions:
            self.keys.append(item['AttributeName'])

    def __get_keys(self, data):
        keys = set([])
        for item in data:
            keys = keys.union(set(item))
        return keys

    def __is_none(self, value):
        None_List = ['None', 'none', 'False', 'false', 'No', 'no', None, False, ["None"], ["False"]]

        if not value:
            return True
        elif value in None_List:
            return True

        return False

    def __check_dynamo_reserved_words(self, k):
        # reserved_words_list = ["ABORT", "ABSOLUTE", "ACTION", "ADD", "AFTER", "AGENT", "AGGREGATE", "ALL", "ALLOCATE", "ALTER", "ANALYZE", "AND", "ANY", "ARCHIVE", "ARE", "ARRAY", "AS", "ASC", "ASCII", "ASENSITIVE", "ASSERTION", "ASYMMETRIC", "AT", "ATOMIC", "ATTACH", "ATTRIBUTE", "AUTH", "AUTHORIZATION", "AUTHORIZE", "AUTO", "AVG", "BACK", "BACKUP", "BASE", "BATCH", "BEFORE", "BEGIN", "BETWEEN", "BIGINT", "BINARY", "BIT", "BLOB", "BLOCK", "BOOLEAN", "BOTH", "BREADTH", "BUCKET", "BULK", "BY", "BYTE", "CALL", "CALLED", "CALLING", "CAPACITY", "CASCADE", "CASCADED", "CASE", "CAST", "CATALOG", "CHAR", "CHARACTER", "CHECK", "CLASS", "CLOB", "CLOSE", "CLUSTER", "CLUSTERED", "CLUSTERING", "CLUSTERS", "COALESCE", "COLLATE", "COLLATION", "COLLECTION", "COLUMN", "COLUMNS", "COMBINE", "COMMENT", "COMMIT", "COMPACT", "COMPILE", "COMPRESS", "CONDITION", "CONFLICT", "CONNECT", "CONNECTION", "CONSISTENCY", "CONSISTENT", "CONSTRAINT", "CONSTRAINTS", "CONSTRUCTOR", "CONSUMED", "CONTINUE", "CONVERT", "COPY", "CORRESPONDING", "COUNT", "COUNTER", "CREATE", "CROSS", "CUBE", "CURRENT", "CURSOR", "CYCLE", "DATA", "DATABASE", "DATE", "DATETIME", "DAY", "DEALLOCATE", "DEC", "DECIMAL", "DECLARE", "DEFAULT", "DEFERRABLE", "DEFERRED", "DEFINE", "DEFINED", "DEFINITION", "DELETE", "DELIMITED", "DEPTH", "DEREF", "DESC", "DESCRIBE", "DESCRIPTOR", "DETACH", "DETERMINISTIC", "DIAGNOSTICS", "DIRECTORIES", "DISABLE", "DISCONNECT", "DISTINCT", "DISTRIBUTE", "DO", "DOMAIN", "DOUBLE", "DROP", "DUMP", "DURATION", "DYNAMIC", "EACH", "ELEMENT", "ELSE", "ELSEIF", "EMPTY", "ENABLE", "END", "EQUAL", "EQUALS", "ERROR", "ESCAPE", "ESCAPED", "EVAL", "EVALUATE", "EXCEEDED", "EXCEPT", "EXCEPTION", "EXCEPTIONS", "EXCLUSIVE", "EXEC", "EXECUTE", "EXISTS", "EXIT", "EXPLAIN", "EXPLODE", "EXPORT", "EXPRESSION", "EXTENDED", "EXTERNAL", "EXTRACT", "FAIL", "FALSE", "FAMILY", "FETCH", "FIELDS", "FILE", "FILTER", "FILTERING", "FINAL", "FINISH", "FIRST", "FIXED", "FLATTERN", "FLOAT", "FOR", "FORCE", "FOREIGN", "FORMAT", "FORWARD", "FOUND", "FREE", "FROM", "FULL", "FUNCTION", "FUNCTIONS", "GENERAL", "GENERATE", "GET", "GLOB", "GLOBAL", "GO", "GOTO", "GRANT", "GREATER", "GROUP", "GROUPING", "HANDLER", "HASH", "HAVE", "HAVING", "HEAP", "HIDDEN", "HOLD", "HOUR", "IDENTIFIED", "IDENTITY", "IF", "IGNORE", "IMMEDIATE", "IMPORT", "IN", "INCLUDING", "INCLUSIVE", "INCREMENT", "INCREMENTAL", "INDEX", "INDEXED", "INDEXES", "INDICATOR", "INFINITE", "INITIALLY", "INLINE", "INNER", "INNTER", "INOUT", "INPUT", "INSENSITIVE", "INSERT", "INSTEAD", "INT", "INTEGER", "INTERSECT", "INTERVAL", "INTO", "INVALIDATE", "IS", "ISOLATION", "ITEM", "ITEMS", "ITERATE", "JOIN", "KEY", "KEYS", "LAG", "LANGUAGE", "LARGE", "LAST", "LATERAL", "LEAD", "LEADING", "LEAVE", "LEFT", "LENGTH", "LESS", "LEVEL", "LIKE", "LIMIT", "LIMITED", "LINES", "LIST", "LOAD", "LOCAL", "LOCALTIME", "LOCALTIMESTAMP", "LOCATION", "LOCATOR", "LOCK", "LOCKS", "LOG", "LOGED", "LONG", "LOOP", "LOWER", "MAP", "MATCH", "MATERIALIZED", "MAX", "MAXLEN", "MEMBER", "MERGE", "METHOD", "METRICS", "MIN", "MINUS", "MINUTE", "MISSING", "MOD", "MODE", "MODIFIES", "MODIFY", "MODULE", "MONTH", "MULTI", "MULTISET", "NAME", "NAMES", "NATIONAL", "NATURAL", "NCHAR", "NCLOB", "NEW", "NEXT", "NO", "NONE", "NOT", "NULL", "NULLIF", "NUMBER", "NUMERIC", "OBJECT", "OF", "OFFLINE", "OFFSET", "OLD", "ON", "ONLINE", "ONLY", "OPAQUE", "OPEN", "OPERATOR", "OPTION", "OR", "ORDER", "ORDINALITY", "OTHER", "OTHERS", "OUT", "OUTER", "OUTPUT", "OVER", "OVERLAPS", "OVERRIDE", "OWNER", "PAD", "PARALLEL", "PARAMETER", "PARAMETERS", "PARTIAL", "PARTITION", "PARTITIONED", "PARTITIONS", "PATH", "PERCENT", "PERCENTILE", "PERMISSION", "PERMISSIONS", "PIPE", "PIPELINED", "PLAN", "POOL", "POSITION", "PRECISION", "PREPARE", "PRESERVE", "PRIMARY", "PRIOR", "PRIVATE", "PRIVILEGES", "PROCEDURE", "PROCESSED", "PROJECT", "PROJECTION", "PROPERTY", "PROVISIONING", "PUBLIC", "PUT", "QUERY", "QUIT", "QUORUM", "RAISE", "RANDOM", "RANGE", "RANK", "RAW", "READ", "READS", "REAL", "REBUILD", "RECORD", "RECURSIVE", "REDUCE", "REF", "REFERENCE", "REFERENCES", "REFERENCING", "REGEXP", "REGION", "REINDEX", "RELATIVE", "RELEASE", "REMAINDER", "RENAME", "REPEAT", "REPLACE", "REQUEST", "RESET", "RESIGNAL", "RESOURCE", "RESPONSE", "RESTORE", "RESTRICT", "RESULT", "RETURN", "RETURNING", "RETURNS", "REVERSE", "REVOKE", "RIGHT", "ROLE", "ROLES", "ROLLBACK", "ROLLUP", "ROUTINE", "ROW", "ROWS", "RULE", "RULES", "SAMPLE", "SATISFIES", "SAVE", "SAVEPOINT", "SCAN", "SCHEMA", "SCOPE", "SCROLL", "SEARCH", "SECOND", "SECTION", "SEGMENT", "SEGMENTS", "SELECT", "SELF", "SEMI", "SENSITIVE", "SEPARATE", "SEQUENCE", "SERIALIZABLE", "SESSION", "SET", "SETS", "SHARD", "SHARE", "SHARED", "SHORT", "SHOW", "SIGNAL", "SIMILAR", "SIZE", "SKEWED", "SMALLINT", "SNAPSHOT", "SOME", "SOURCE", "SPACE", "SPACES", "SPARSE", "SPECIFIC", "SPECIFICTYPE", "SPLIT", "SQL", "SQLCODE", "SQLERROR", "SQLEXCEPTION", "SQLSTATE", "SQLWARNING", "START", "STATE", "STATIC", "STATUS", "STORAGE", "STORE", "STORED", "STREAM", "STRING", "STRUCT", "STYLE", "SUB", "SUBMULTISET", "SUBPARTITION", "SUBSTRING", "SUBTYPE", "SUM", "SUPER", "SYMMETRIC", "SYNONYM", "SYSTEM", "TABLE", "TABLESAMPLE", "TEMP", "TEMPORARY", "TERMINATED", "TEXT", "THAN", "THEN", "THROUGHPUT", "TIME", "TIMESTAMP", "TIMEZONE", "TINYINT", "TO", "TOKEN", "TOTAL", "TOUCH", "TRAILING", "TRANSACTION", "TRANSFORM", "TRANSLATE", "TRANSLATION", "TREAT", "TRIGGER", "TRIM", "TRUE", "TRUNCATE", "TTL", "TUPLE", "TYPE", "UNDER", "UNDO", "UNION", "UNIQUE", "UNIT", "UNKNOWN", "UNLOGGED", "UNNEST", "UNPROCESSED", "UNSIGNED", "UNTIL", "UPDATE", "UPPER", "URL", "USAGE", "USE", "USER", "USERS", "USING", "UUID", "VACUUM", "VALUE", "VALUED", "VALUES", "VARCHAR", "VARIABLE", "VARIANCE", "VARINT", "VARYING", "VIEW", "VIEWS", "VIRTUAL", "VOID", "WAIT", "WHEN", "WHENEVER", "WHERE", "WHILE", "WINDOW", "WITH", "WITHIN", "WITHOUT", "WORK", "WRAPPED", "WRITE", "YEAR", "ZONE"]
        # if k.upper() in reserved_words_list:
        #     print(f"You have a reserved word: {k}")
        #     return True
        # elif "." in k:
        #     print(f"{k} has a period and it will not evaluate properly in later upserts")
        #     return True

        return False

    def __standardize(self, data, filter_empty=False):

        if not isinstance(data, dict):
            print("wrong data type for dynamodb")
            return None

        for k, v in data.items():
            if self.__is_none(v):
                data[k] = None
            if isinstance(v, float):
                data[k] = Decimal(str(v))
            if self.__check_dynamo_reserved_words(k):
                raise ValueError("DynamoDB Reserved Words Key Error. Change your keys.")


        if filter_empty:
            # filter out all empty values from dict()
            return { k:v for k,v in data.items() if v is not None }

        return data

    def import_csv_file(self, file_name, delimiter=','):
        print(f'>> import_csv_file({file_name})')
        items = []

        with open(file_name) as csv_file:
            items = list(map(self.__standardize, csv.DictReader(csv_file)))

        with self.table.batch_writer() as batch:
            for item in items:
                batch.put_item(Item=item)

        print("\nSucccessfully written {} record(s) to {}\n".format(len(items), self.table_name))
        return True

    def import_json_file(self, file_name):
        click.echo("NOT IMPLEMENTED YET ... SORRY")
        pass

    def import_file(self, file_name, format):
        if "csv" == format:
            return self.import_csv_file(file_name)
        elif "json" == format:
            return self.import_json_file(file_name)
        else:
            raise Exception(f'Unknown format: {format}')


import os
import click
import boto3
from dyndb import DynamoExporter, DynamoImporter

@click.group()
def main():
    pass

@main.command('export')
@click.option('-t', '--table', help='DynamoDB table to export', required=True)
@click.option('-r', '--region', help='AWS region', default='eu-central-1')
@click.option('-f', '--format', help='Format file [csv/json].', type=click.Choice(['json','csv']), default='csv')
@click.option('-o', '--output', help='Output directory', type=click.Path(), default=os.path.join(os.getcwd(),'out'))
def exp(table, region, format, output):
    """
    Export data from DynamoDB table
    """
    clz = DynamoExporter(table_name=table, region=region)

    data = clz.get_data()

    if not os.path.isdir(output):
        os.makedirs(output)

    if data:        
        if format == 'json':
            clz.write_to_json_file(data, os.path.join(output, '{}.json'.format(table)))
        else:
            clz.write_to_csv_file(data, os.path.join(output, '{}.csv'.format(table)))
    else:
        click.echo('No data to export')

@main.command('import')
@click.option('-t', '--table', help='DynamoDB table to export')
@click.option('-r', '--region', help='AWS region', default='eu-central-1')
@click.option('-f', '--file', help='File', type=click.Path(), required=True)
def imp(table, region, file):
    """
    Import data into DynamoDB table
    """
    base = os.path.splitext(os.path.basename(file))
    format = base[1][1:]

    if not table:
        table = base[0]

    clz = DynamoImporter(table_name=table, region=region)

    clz.import_file(file, format=format)

@main.command()
@click.option('-r', '--region', help='AWS region', default='eu-central-1')
@click.option('-l', '--limit', help='A maximum number of table names to return', default=100)
def list(region, limit):
    """
    List all available DynamoDB tables
    """
    client = boto3.client('dynamodb', region_name=region)

    response = client.list_tables(
        Limit = limit
    )

    click.echo('\nFound {} DynamoDB tables in region {}.\n'.format(len(response['TableNames']),region))

    for t in response['TableNames']:
        click.echo(' - {}'.format(t))


@main.command('truncate')
@click.option('-t', '--table', help='DynamoDB table to export', required=True)
@click.option('-r', '--region', help='AWS region', default='eu-central-1')
def trunc(table, region):
    """
    Truncate all items from DynamoDB table
    """
    click.confirm("About to truncate table '{}'. Are you sure?".format(table), abort=True)

    resource = boto3.resource('dynamodb', region_name=region)

    table = resource.Table(table)

    keys = []

    for item in table.attribute_definitions:
        keys.append(item['AttributeName'])

    scan = table.scan()

    with table.batch_writer() as batch:
        for item in scan['Items']:
            batch.delete_item(
                Key = { k: v for k, v in item.items() if k in keys}
            )


from .exp import DynamoExporter
from .imp import DynamoImporter

import boto3
from botocore.exceptions import ClientError

def ddb_exists(table_name, region='eu-central-1'):
    try:
        boto3.client('dynamodb', region_name=region).describe_table(TableName=table_name)

        return True

    except ClientError as ce:

        if ce.response['Error']['Code'] == 'ResourceNotFoundException':
            return False
        else:
            raise ce


import json
import csv
import boto3

import dyndb

class DynamoExporter():
    def __init__(self, table_name, region):
        self.resource = boto3.resource('dynamodb', region_name=region)
        self.table_name = table_name
        self.region = region
        self.table = self.resource.Table(table_name)

    def __get_keys(self, data):
        keys = set([])
        for item in data:
            keys = keys.union(set(item))
        return keys

    def __convert_rawdata_to_stringvalue(self, data):
        """
        Convert raw data to string value.
        :param data: List of dictionary
        :return: String value.
        """
        ret = []
        for item in data:
            obj = {}
            for k, v in item.items():
                obj[k] = str(v)
            ret.append(obj)
        return ret

    def get_data(self):
        keys = []

        for item in self.table.attribute_definitions:
            keys.append(item['AttributeName'])

        keys_set = set(keys)

        item_count = self.table.item_count

        raw_data = self.table.scan()
        if not raw_data:
            return None

        items = raw_data['Items']
        fieldnames = set([]).union(self.__get_keys(items))

        cur_total = (len(items) + raw_data['Count'])
        if cur_total > item_count:
            percents = 99.99
        else:
            percents = cur_total * 100 / item_count

        print("{} records ..... {:02.0f}%".format(raw_data['Count'], percents), end='\r')

        while raw_data.get('LastEvaluatedKey'):
            print('Downloading ', end='')
            raw_data = self.table.scan(ExclusiveStartKey=raw_data['LastEvaluatedKey'])
            items.extend(raw_data['Items'])
            fieldnames = fieldnames.union(self.__get_keys(items))
            cur_total = (len(items) + raw_data['Count'])
            if cur_total > item_count:
                percents = 99.99
            else:
                percents = cur_total * 100 / item_count

            print("{} records ..... {:02.0f}%".format(raw_data['Count'], percents), end='\r')

        print("\nTotal downloaded records: {}".format(len(items)))

        for fieldname in fieldnames:
            if fieldname not in keys:
                keys.append(fieldname)

        return {'items': items, 'keys': keys}

    def write_to_json_file(self, data, filename):
        """
        Write to a json file
        :param data: Dictionary
        :param filename: output file name.
        :return: None
        """
        if data is None:
            return

        with open(filename, 'w') as f:
            f.write(json.dumps(self.__convert_rawdata_to_stringvalue(data['items'])))

        print("\nSucccessfully exported {} item(s) to '{}'\n".format(len(data['items']), filename))
        return True

    def write_to_csv_file(self, data, filename):
        """
        Write to a csv file.
        :param data:
        :param filename:
        :return: True
        """
        if data is None:
            return

        with open(filename, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, delimiter=',', fieldnames=data['keys'], quotechar='"')
            writer.writeheader()
            writer.writerows(data['items'])

        print("\nSucccessfully exported {} item(s) to '{}'\n".format(len(data['items']), filename))
        return True


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dyndb']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.12.26,<2.0.0', 'click>=7.1.1,<8.0.0']

entry_points = \
{'console_scripts': ['dyndb = dyndb.cli:main']}

setup_kwargs = {
    'name': 'dyndb',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Pascal Prins',
    'author_email': 'pascal.prins@foobar-it.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*',
}


setup(**setup_kwargs)
